
#pragma hdrstop
#include <condefs.h>


//---------------------------------------------------------------------------
/* mc.c Marching Cubes implementation for lab1
   Mike Chupa
   CS 8413: Visualization
   Instructor: Dr. Raghu Machiraju
   Copyright (c) 1998 Michael A. Chupa Permission is expressly granted
   to use this code in any non-commercial work, provided that this
   notice is preserved. However, be aware that the Marching Cubes
   algorithm is patented.
 
   Modified by: Andrew Le Couteur Bisson - Feb 2003
   Now produces POVRay Mesh files from Povray density files (df3)
*/

/* usage: mc dataset threshold_value [output_file] [a b c]*/
/*        where   dataset = path to a scalar field dataset (byte)
           iDim, jDim, kDim = dataset dimensions
    threshold_value  = isosurface threshold value
    output_file      = path to inventor scene graph output (dataset.iv)
    a b c            = aspect ratio of input dataset (1:1:1 default)
*/

#include "mc.h"

int iDim = 0;              /* dataset dimensions             */
int jDim = 0;
int kDim = 0;

int main(int argc, char **argv)
{

  FILE *in;                 /* input dataset file             */
  FILE *out;                /* output file for Inventor scene */
  float threshold;          /* threshold value for isosurface */

  datum *dataset;            /* the scalar field               */
  Layer *layer;              /* scalar field data and edges    */
  isosurface *iso;                /* vertices, normals and triangles*/

  int datasize;               /* size of dataset in bytes       */
  int bytesRead;              /* actual number of bytes read    */
  int i, j, k, e;
  long cpuTime;


  if (argc != 3)
  {
    fprintf(stderr, "usage: mc datafile threshold\n");
    exit(1);
  }

  /* parse command line */
  threshold = atof(argv[2]);

  /* open input dataset */
  if ((in = fopen(argv[1], "rb")) == 0)
  {
    perror(argv[1]);
    exit(1);
  }
  if (((bytesRead = fread((char *) & iDim + 1, 1, 1, in)) != 1) ||
      ((bytesRead = fread((char *) & iDim, 1, 1, in)) != 1) ||
      ((bytesRead = fread((char *) & jDim + 1, 1, 1, in)) != 1) ||
      ((bytesRead = fread((char *) & jDim, 1, 1, in)) != 1) ||
      ((bytesRead = fread((char *) & kDim + 1, 1, 1, in)) != 1) ||
      ((bytesRead = fread((char *) & kDim, 1, 1, in)) != 1))
  {
    fprintf(stderr, "%s: failed to read dataset\n", argv[0]);
    fprintf(stderr, "Bytes requested: 1, Bytes read: %i\n", bytesRead);
    exit(1);
  }

  /* create the output file with name of */
  /* input dataset with a .inc extension tacked on...   */
  if ((out = fopen(strcat(argv[1], ".inc"), "w")) == NULL)
  {
    perror(strcat(argv[1], ".inc"));
    exit(1);
  }

  /* now that output file is open, write the POV header */
  writePOVHeader(out);

  /* allocate storage and read the data */

  datasize = iDim * jDim * kDim * sizeof(datum);
  dataset = (datum *) malloc(datasize);
  if (dataset == NULL)
  {
    fprintf(stderr, "%s: failed to allocate dataset storage\n", argv[0]);
    exit(1);
  }

  if ((bytesRead = fread((void *)dataset, 1, datasize, in)) != datasize)
  {
    fprintf(stderr, "%s: failed to read dataset\n", argv[0]);
    fprintf(stderr, "Bytes requested: %i, Bytes read: %i\n", datasize, bytesRead);
    exit(1);
  }

  layer = (Layer *) malloc(sizeof(Layer));
  if (layer == NULL)
  {
    fprintf(stderr, "Failed to allocate layer\n");
    exit(1);
  }

  /* initially, botData points to the start of dataset, and */
  /* topData points to the next layer */
  layer->botData = dataset;
  layer->topData = dataset + OFFSET(0, 0, 1, iDim, jDim);
  /* allocate storage to hold the indexing tags for edges in the layer */
  layer->edges = (long *) malloc((iDim - 1) * (jDim - 1) * 12 * sizeof(long));
  if (layer->edges == NULL)
  {
    fprintf(stderr, "Failed to allocate layer edges\n");
    exit(1);
  }
  /* initialize the edge table to EMPTY */
  for (i = 0; i < (iDim - 1)*(jDim - 1)*12; i++)
  {
    *(layer->edges + i) = EMPTY_EDGE;
  }
#ifdef _DEBUG2_
  for (i = 0; i < iDim - 1; i++)
  {
    for (j = 1; j < jDim; j++)
    {
      fprintf(stderr, "Edges for cell %i, %i\n",
              i, j - 1);
      for (e = 0; e < 12; e++)
      {
        fprintf(stderr, "%i ",
                *(layer->edges + EDGE_OFFSET(e, i, j, iDim)));
      }
      fprintf(stderr, "\n\n");
    }
  }
#endif


  /* allocate the storage for the isosurface and initialize */
  iso = (isosurface *) malloc(sizeof(isosurface));
  for (i = 0; i < 3; i++)
  {
    iso->vertices[i] = NULL;
    iso->normals[i] = NULL;
    iso->triangles[i] = NULL;
    iso->nVertices = 0;
    iso->nTriangles = 0;
  }

  /* now do marching cubes on the first layer of cells */
#ifdef _DEBUG_
  fprintf(stderr, "Running marching cubes on layer   0: ");
#endif

  march(dataset, layer, 0, threshold, iso);

#ifdef _DEBUG2_

  for (i = 0; i < iDim - 1; i++)
  {
    for (j = 1; j < jDim; j++)
    {
      fprintf(stderr, "Edges for cell %i, %i\n",
              i, j - 1);
      for (e = 0; e < 12; e++)
      {
        fprintf(stderr, "%i ",
                *(layer->edges + EDGE_OFFSET(e, i, j, iDim)));
      }
      fprintf(stderr, "\n\n");
    }
  }
#endif


  /* now do the remaining layers */
  for (k = 1; k < kDim - 1; k++)
  {
    /* update the layer for this iteration */
    layer->botData = layer->topData;
    /* now topData points to next layer of scalar data*/
    layer->topData = dataset + OFFSET(0, 0, k + 1, iDim, jDim);

    /* percolate the last layer's top edges to this layer's bottom edges */
    for (i = 0; i < iDim - 1; i++)
    {
      for (j = 1; j < jDim; j++)
      {
        for (e = 0; e < 4; e++)
        {
          *(layer->edges + EDGE_OFFSET(e, i, j, iDim)) =
            *(layer->edges + EDGE_OFFSET(e + 4, i, j, iDim));
        }
        /* reinitialize all of the remaining edges */
        for (e = 4; e < 12; e++)
        {
          *(layer->edges + EDGE_OFFSET(e, i, j, iDim)) = EMPTY_EDGE;
        }
      }
    }



#ifdef _DEBUG_
    fprintf(stderr, "Running marching cubes on layer %3i: ", k);
#endif

    march(dataset, layer, k, threshold, iso);


#ifdef _DEBUG2_

    for (i = 0; i < iDim - 1; i++)
    {
      for (j = 1; j < jDim; j++)
      {
        fprintf(stderr, "Edges for cell %i, %i\n",
                i, j - 1);
        for (e = 0; e < 12; e++)
        {
          fprintf(stderr, "%i ",
                  *(layer->edges + EDGE_OFFSET(e, i, j, iDim)));
        }
        fprintf(stderr, "\n\n");
      }
    }
#endif

  }

  /* print some stats */
  cpuTime = clock();
  fprintf(stderr, "Isosurface contains\n\t%i vertices and\n\t%i triangles.\n\n",
          iso->nVertices, iso->nTriangles);
  fprintf(stderr, "CPU Time used: %g seconds\n", cpuTime * 1E-6);

#ifdef _DEBUG_

  fprintf(stderr, "Writing out isosurface to POVRay mesh2 file...\n");
#endif

  writeVertices(iso, out);
  writeNormals(iso, out);
  writeTriangles(iso, out);

  /* clean up */
  fclose(out);
  fclose(in);
  free(dataset);
  free(layer->edges);
  free(layer);
  for (i = 0; i < 3; i++)
  {
    free(iso->vertices[i]);
    free(iso->normals[i]);
    free(iso->triangles[i]);
  }
  free(iso);
}


void march(datum *dataset, Layer *layer, int k, float threshold, isosurface *iso)
{

  isosurface *layerIso;          /* portion of isosurface in layer   */

  float *fPtr;
  long *lPtr;
  datum cell[8];           /* grid values at 8 cell vertices   */
  long cellVerts[12];     /* verts at the 12 edges in a cell  */
  int i, j, ii, jj, e;
  long vertSize;          /* allocation for verts and normals */
  long triSize;           /* allocation for triangles         */
  unsigned char cellIndex;         /* index for edgeTable lookup       */



  /* initialize the cellVerts table */
  for (i = 0; i < 12; i++)
  {
    cellVerts[i] = EMPTY_EDGE;
  }

  /* allocate and initialize storage for this layer's portion of isosurface */
  layerIso = (isosurface *) malloc(sizeof(isosurface));

  /* we have a maximum of 12 verts and 5 triangles per cell*/
  vertSize = (iDim - 1) * (jDim - 1) * 12 * sizeof(float);
  triSize = (iDim - 1) * (jDim - 1) * 5 * sizeof(int);
  /* allocate storage for vertices, normals and triangles */
  for (i = 0; i < 3; i++)
  {
    layerIso->vertices[i] = (float *) malloc(vertSize);
    layerIso->normals[i] = (float *) malloc(vertSize);
    layerIso->triangles[i] = (long *) malloc(triSize );

    if ((layerIso->vertices[i] == NULL) ||
        (layerIso->normals[i] == NULL) ||
        (layerIso->triangles[i] == NULL))
    {
      fprintf(stderr, "march: failed to allocate layer isosurface\n");
      exit(1);
    }
  }
  /* init counts */
  layerIso->nVertices = 0;
  layerIso->nTriangles = 0;

  /* for each cell in the grid... */
  for (i = 0; i < iDim - 1; i++)
  {
    for (j = 1; j < jDim; j++)
    {
      /* do the marching cubes thing */
      /* fill the cell */
      cell[0] = *(layer->botData + j * iDim + i);
      cell[1] = *(layer->botData + j * iDim + i + 1);
      cell[2] = *(layer->botData + (j - 1) * iDim + i + 1);
      cell[3] = *(layer->botData + (j - 1) * iDim + i);
      cell[4] = *(layer->topData + j * iDim + i);
      cell[5] = *(layer->topData + j * iDim + i + 1);
      cell[6] = *(layer->topData + (j - 1) * iDim + i + 1);
      cell[7] = *(layer->topData + (j - 1) * iDim + i);

      /* compute the index for the edge intersections */
      cellIndex = 0;
      if (cell[0] < threshold)
        cellIndex |= 1;
      if (cell[1] < threshold)
        cellIndex |= 2;
      if (cell[2] < threshold)
        cellIndex |= 4;
      if (cell[3] < threshold)
        cellIndex |= 8;
      if (cell[4] < threshold)
        cellIndex |= 16;
      if (cell[5] < threshold)
        cellIndex |= 32;
      if (cell[6] < threshold)
        cellIndex |= 64;
      if (cell[7] < threshold)
        cellIndex |= 128;

      /* get the coordinates for the vertices */
      /* compute the triangulation */
      /* interpolate the normals */
      if (edgeTable[cellIndex] & 1)
      {
        if (*(layer->edges + EDGE_OFFSET(0, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[0] = iso->nVertices +
                         makeVertex(0, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[0] = *(layer->edges + EDGE_OFFSET(0, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 2)
      {
        if (*(layer->edges + EDGE_OFFSET(1, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[1] = iso->nVertices +
                         makeVertex(1, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[1] = *(layer->edges + EDGE_OFFSET(1, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 4)
      {
        if (*(layer->edges + EDGE_OFFSET(2, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[2] = iso->nVertices +
                         makeVertex(2, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[2] = *(layer->edges + EDGE_OFFSET(2, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 8)
      {
        if (*(layer->edges + EDGE_OFFSET(3, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[3] = iso->nVertices +
                         makeVertex(3, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[3] = *(layer->edges + EDGE_OFFSET(3, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 16)
      {
        if (*(layer->edges + EDGE_OFFSET(4, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[4] = iso->nVertices +
                         makeVertex(4, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[4] = *(layer->edges + EDGE_OFFSET(4, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 32)
      {
        if (*(layer->edges + EDGE_OFFSET(5, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[5] = iso->nVertices +
                         makeVertex(5, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[5] = *(layer->edges + EDGE_OFFSET(5, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 64)
      {
        if (*(layer->edges + EDGE_OFFSET(6, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[6] = iso->nVertices +
                         makeVertex(6, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[6] = *(layer->edges + EDGE_OFFSET(6, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 128)
      {
        if (*(layer->edges + EDGE_OFFSET(7, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[7] = iso->nVertices +
                         makeVertex(7, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[7] = *(layer->edges + EDGE_OFFSET(7, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 256)
      {
        if (*(layer->edges + EDGE_OFFSET(8, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[8] = iso->nVertices +
                         makeVertex(8, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[8] = *(layer->edges + EDGE_OFFSET(8, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 512)
      {
        if (*(layer->edges + EDGE_OFFSET(9, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[9] = iso->nVertices +
                         makeVertex(9, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[9] = *(layer->edges + EDGE_OFFSET(9, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 1024)
      {
        if (*(layer->edges + EDGE_OFFSET(10, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[10] = iso->nVertices +
                          makeVertex(10, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[10] = *(layer->edges + EDGE_OFFSET(10, i, j, iDim));
        }
      }
      if (edgeTable[cellIndex] & 2048)
      {
        if (*(layer->edges + EDGE_OFFSET(11, i, j, iDim)) == EMPTY_EDGE)
        {
          cellVerts[11] = iso->nVertices +
                          makeVertex(11, i, j, k, threshold, dataset, layerIso);
        }
        else
        {
          cellVerts[11] = *(layer->edges + EDGE_OFFSET(11, i, j, iDim));
        }
      }

      /* put the cellVerts tags into this cell's layer->edges table */
      for (e = 0; e < 12; e++)
      {
        if (cellVerts[e] != EMPTY_EDGE)
        {
          *(layer->edges + EDGE_OFFSET(e, i, j, iDim)) = cellVerts[e];
        }
      }

      /* now propagate the vertex/normal tags to the adjacent cells to */
      /* the right and behind in this layer. */
      if (i < iDim - 2)
      { /* we should propagate to the right */
        *(layer->edges + EDGE_OFFSET( 3, i + 1, j, iDim)) = cellVerts[1];
        *(layer->edges + EDGE_OFFSET( 7, i + 1, j, iDim)) = cellVerts[5];
        *(layer->edges + EDGE_OFFSET( 8, i + 1, j, iDim)) = cellVerts[9];
        *(layer->edges + EDGE_OFFSET(11, i + 1, j, iDim)) = cellVerts[10];
      }
      if (j < jDim - 1)
      { /* we should propagate to the rear */
        *(layer->edges + EDGE_OFFSET( 2, i, j + 1, iDim)) = cellVerts[0];
        *(layer->edges + EDGE_OFFSET( 6, i, j + 1, iDim)) = cellVerts[4];
        *(layer->edges + EDGE_OFFSET(11, i, j + 1, iDim)) = cellVerts[8];
        *(layer->edges + EDGE_OFFSET(10, i, j + 1, iDim)) = cellVerts[9];
      }


      /* compute the triangulation */
      ii = 0;
      while (triTable[cellIndex][ii] != -1)
      {
        for (jj = 0; jj < 3; jj++)
        {
          *(layerIso->triangles[jj] + layerIso->nTriangles) = cellVerts[triTable[cellIndex][ii++]];
        }
        layerIso->nTriangles++;
      }

    }
  }

  /* now that we have the new verts, normals and triangles */
  /* calculated, we need to add them to the master lists   */

#ifdef _DEBUG_
  fprintf(stderr, "%5i verts, %5i triangles\n", layerIso->nVertices, layerIso->nTriangles);
#endif

  /* if new verts, expand the storage for the master lists */
  if (layerIso->nVertices > 0)
  {
    for (i = 0; i < 3; i++)
    {
      iso->vertices[i] = (float *) realloc((void *) iso->vertices[i],
                                           (iso->nVertices + layerIso->nVertices) * sizeof(float));
      iso->normals[i] = (float *) realloc((void *) iso->normals[i],
                                          (iso->nVertices + layerIso->nVertices) * sizeof(float));
      iso->triangles[i] = (long *) realloc((void *) iso->triangles[i],
                                           (iso->nTriangles + layerIso->nTriangles) * sizeof(long));

      if ((iso->vertices[i] == NULL) ||
          (iso->normals[i] == NULL) ||
          (iso->triangles[i] == NULL))
      {
        fprintf(stderr, "realloc failed\n");
        exit(1);
      }

      /* tack on the new data to the master lists */
      fPtr = iso->vertices[i] + iso->nVertices;
      memcpy((void *) fPtr, layerIso->vertices[i], layerIso->nVertices * sizeof(float));
      fPtr = iso->normals[i] + iso->nVertices;
      memcpy((void *) fPtr, layerIso->normals[i], layerIso->nVertices * sizeof(float));
      lPtr = iso->triangles[i] + iso->nTriangles;
      memcpy((void *) lPtr, layerIso->triangles[i], layerIso->nTriangles * sizeof(long));

    }
    /* update the master list counts */
    iso->nVertices += layerIso->nVertices;
    iso->nTriangles += layerIso->nTriangles;

#ifdef _DEBUG_

    fprintf(stderr, "\t\ttotals: %i verts and %i triangles\n",
            iso->nVertices, iso->nTriangles);
#endif

  }

  /* toss the storage for this run of march */
  for (i = 0; i < 3; i++)
  {
    free(layerIso->vertices[i]);
    free(layerIso->normals[i]);
    free(layerIso->triangles[i]);
  }
  free(layerIso);
}


long makeVertex(int whichEdge, int i, int j, int k,
                float threshold, datum *dataset, isosurface *layerIso)
{

  int from[3]; /* first grid vertex                              */
  int to[3]; /* second grid vertex                             */
  float normFrom[3]; /* normal at from vertex                          */
  float normTo[3]; /* normal at the to vertex                        */
  float v[3]; /* the interpolated vertex                        */
  float n[3]; /* and its normal                                 */
  float d;    /* relative distance along edge for interpolation */
  float len;    /* normal length for normalization                */
  int ii;

  switch (whichEdge)
  {
    case 0:
      from[0] = i;
      from[1] = j;
      from[2] = k;
      to[0] = i + 1;
      to[1] = j;
      to[2] = k;
      break;
    case 1:
      from[0] = i + 1;
      from[1] = j;
      from[2] = k;
      to[0] = i + 1;
      to[1] = j - 1;
      to[2] = k;
      break;
    case 2:
      from[0] = i + 1;
      from[1] = j - 1;
      from[2] = k;
      to[0] = i;
      to[1] = j - 1;
      to[2] = k;
      break;
    case 3:
      from[0] = i;
      from[1] = j - 1;
      from[2] = k;
      to[0] = i;
      to[1] = j;
      to[2] = k;
      break;
    case 4:
      from[0] = i;
      from[1] = j;
      from[2] = k + 1;
      to[0] = i + 1;
      to[1] = j;
      to[2] = k + 1;
      break;
    case 5:
      from[0] = i + 1;
      from[1] = j;
      from[2] = k + 1;
      to[0] = i + 1;
      to[1] = j - 1;
      to[2] = k + 1;
      break;
    case 6:
      from[0] = i + 1;
      from[1] = j - 1;
      from[2] = k + 1;
      to[0] = i;
      to[1] = j - 1;
      to[2] = k + 1;
      break;
    case 7:
      from[0] = i;
      from[1] = j - 1;
      from[2] = k + 1;
      to[0] = i;
      to[1] = j;
      to[2] = k + 1;
      break;
    case 8:
      from[0] = i;
      from[1] = j;
      from[2] = k;
      to[0] = i;
      to[1] = j;
      to[2] = k + 1;
      break;
    case 9:
      from[0] = i + 1;
      from[1] = j;
      from[2] = k;
      to[0] = i + 1;
      to[1] = j;
      to[2] = k + 1;
      break;
    case 10:
      from[0] = i + 1;
      from[1] = j - 1;
      from[2] = k;
      to[0] = i + 1;
      to[1] = j - 1;
      to[2] = k + 1;
      break;
    case 11:
      from[0] = i;
      from[1] = j - 1;
      from[2] = k;
      to[0] = i;
      to[1] = j - 1;
      to[2] = k + 1;
      break;
    default:
      fprintf(stderr, "mc: makeVertex: bad edge index\n");
      exit(1);
      break;
  }

  /* determine the relative distance along edge from->to */
  /* that the isosurface vertex lies */
  d = ( *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)) - threshold) /
      ( *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)) -
        *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)) );
  if (d < EPSILON)
  {
    d = 0.0;
  }
  else if (d > (1 - EPSILON))
  {
    d = 1.0;
  }

  v[0] = from[0] + d * (to[0] - from[0]);
  v[1] = from[1] + d * (to[1] - from[1]);
  v[2] = from[2] + d * (to[2] - from[2]);

  /* now determine the gradients at the endpoints of the edge */
  /* and interpolate the normal for the isosurface vertex     */

  /* the gradients are computed by central differences, except */
  /* on the boundaries of the dataset, where forward or backward */
  /* differencing is used (three point form) */
  if (from[0] == 0)
  { /* on left edge */
    normFrom[0] = 0.5 * ( -3.0 * *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)) +
                          4.0 * *(dataset + OFFSET(from[0] + 1, from[1], from[2], iDim, jDim)) -
                          *(dataset + OFFSET(from[0] + 2, from[1], from[2], iDim, jDim)));
  }
  else if (from[0] == iDim - 1)
  { /* on right edge */
    normFrom[0] = 0.5 * ( *(dataset + OFFSET(from[0] - 2, from[1], from[2], iDim, jDim)) -
                          4.0 * *(dataset + OFFSET(from[0] - 1, from[1], from[2], iDim, jDim)) +
                          3.0 * *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)));
  }
  else
  { /* in the interior */
    normFrom[0] = 0.5 * ( *(dataset + OFFSET(from[0] + 1, from[1], from[2], iDim, jDim)) -
                          *(dataset + OFFSET(from[0] - 1, from[1], from[2], iDim, jDim)));
  }

  if (from[1] == 0)
  { /* on front edge */
    normFrom[1] = 0.5 * ( -3.0 * *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)) +
                          4.0 * *(dataset + OFFSET(from[0], from[1] + 1, from[2], iDim, jDim)) -
                          *(dataset + OFFSET(from[0], from[1] + 2, from[2], iDim, jDim)));
  }
  else if (from[1] == jDim - 1)
  { /* on back edge */
    normFrom[1] = 0.5 * ( *(dataset + OFFSET(from[0], from[1] - 2, from[2], iDim, jDim)) -
                          4.0 * *(dataset + OFFSET(from[0], from[1] - 1, from[2], iDim, jDim)) +
                          3.0 * *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)));
  }
  else
  { /* in the interior */
    normFrom[1] = 0.5 * ( *(dataset + OFFSET(from[0], from[1] + 1, from[2], iDim, jDim)) -
                          *(dataset + OFFSET(from[0], from[1] - 1, from[2], iDim, jDim)));
  }

  if (from[2] == 0)
  { /* on bottom edge */
    normFrom[2] = 0.5 * ( -3.0 * *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)) +
                          4.0 * *(dataset + OFFSET(from[0], from[1], from[2] + 1, iDim, jDim)) -
                          *(dataset + OFFSET(from[0], from[1], from[2] + 2, iDim, jDim)));
  }
  else if (from[2] == kDim - 1)
  { /* on top edge */
    normFrom[2] = 0.5 * ( *(dataset + OFFSET(from[0], from[1], from[2] - 2, iDim, jDim)) -
                          4.0 * *(dataset + OFFSET(from[0], from[1], from[2] - 1, iDim, jDim)) +
                          3.0 * *(dataset + OFFSET(from[0], from[1], from[2], iDim, jDim)));
  }
  else
  { /* in the interior */
    normFrom[2] = 0.5 * ( *(dataset + OFFSET(from[0], from[1], from[2] + 1, iDim, jDim)) -
                          *(dataset + OFFSET(from[0], from[1], from[2] - 1, iDim, jDim)));
  }


  /* normal for to vertex */

  if (to[0] == 0)
  { /* on left edge */
    normTo[0] = 0.5 * ( -3.0 * *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)) +
                        4.0 * *(dataset + OFFSET(to[0] + 1, to[1], to[2], iDim, jDim)) -
                        *(dataset + OFFSET(to[0] + 2, to[1], to[2], iDim, jDim)));
  }
  else if (to[0] == iDim - 1)
  { /* on right edge */
    normTo[0] = 0.5 * ( *(dataset + OFFSET(to[0] - 2, to[1], to[2], iDim, jDim)) -
                        4.0 * *(dataset + OFFSET(to[0] - 1, to[1], to[2], iDim, jDim)) +
                        3.0 * *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)));
  }
  else
  { /* in the interior */
    normTo[0] = 0.5 * ( *(dataset + OFFSET(to[0] + 1, to[1], to[2], iDim, jDim)) -
                        *(dataset + OFFSET(to[0] - 1, to[1], to[2], iDim, jDim)));
  }

  if (to[1] == 0)
  { /* on front edge */
    normTo[1] = 0.5 * ( -3.0 * *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)) +
                        4.0 * *(dataset + OFFSET(to[0], to[1] + 1, to[2], iDim, jDim)) -
                        *(dataset + OFFSET(to[0], to[1] + 2, to[2], iDim, jDim)));
  }
  else if (to[1] == jDim - 1)
  { /* on back edge */
    normTo[1] = 0.5 * ( *(dataset + OFFSET(to[0], to[1] - 2, to[2], iDim, jDim)) -
                        4.0 * *(dataset + OFFSET(to[0], to[1] - 1, to[2], iDim, jDim)) +
                        3.0 * *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)));
  }
  else
  { /* in the interior */
    normTo[1] = 0.5 * ( *(dataset + OFFSET(to[0], to[1] + 1, to[2], iDim, jDim)) -
                        *(dataset + OFFSET(to[0], to[1] - 1, to[2], iDim, jDim)));
  }

  if (to[2] == 0)
  { /* on bottom edge */
    normTo[2] = 0.5 * ( -3.0 * *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)) +
                        4.0 * *(dataset + OFFSET(to[0], to[1], to[2] + 1, iDim, jDim)) -
                        *(dataset + OFFSET(to[0], to[1], to[2] + 2, iDim, jDim)));
  }
  else if (to[2] == kDim - 1)
  { /* on top edge */
    normTo[2] = 0.5 * ( *(dataset + OFFSET(to[0], to[1], to[2] - 2, iDim, jDim)) -
                        4.0 * *(dataset + OFFSET(to[0], to[1], to[2] - 1, iDim, jDim)) +
                        3.0 * *(dataset + OFFSET(to[0], to[1], to[2], iDim, jDim)));
  }
  else
  { /* in the interior */
    normTo[2] = 0.5 * ( *(dataset + OFFSET(to[0], to[1], to[2] + 1, iDim, jDim)) -
                        *(dataset + OFFSET(to[0], to[1], to[2] - 1, iDim, jDim)));
  }

  /* now that we have the normals at the two endpoints, interpolate */

  n[0] = normFrom[0] + d * (normTo[0] - normFrom[0]);
  n[1] = normFrom[1] + d * (normTo[1] - normFrom[1]);
  n[2] = normFrom[2] + d * (normTo[2] - normFrom[2]);

  /* normalize the normal at the isosurface vertex */
  len = sqrt(n[0] * n[0] + n[1] * n[1] + n[2] * n[2]);
  if (len > EPSILON)
  {
    n[0] /= len;
    n[1] /= len;
    n[2] /= len;
  }
  else
  { /* we have to fake this normal */
    n[0] = 1.0;
    n[1] = 0.0;
    n[2] = 0.0;
  }

  /* insert the vertex and normal into the isosurface structure */
  /* compute the tag for this vertex, and return the tag        */

  for (ii = 0; ii < 3; ii++)
  {
    *(layerIso->vertices[ii] + layerIso->nVertices) = v[ii];
    *(layerIso->normals[ii] + layerIso->nVertices) = n[ii];
  }
  layerIso->nVertices++;

  return layerIso->nVertices - 1;

}

void writePOVHeader(FILE *f)
{
  fprintf(f, "mesh2\n");
  fprintf(f, "{\n");

}
void writeVertices(isosurface *iso, FILE *f)
{

  float *fPtr[3]; /* pointer into the vertex list */
  long ii;


  fprintf(f, "  vertex_vectors\n  { ");
  fprintf(f, "  %d,\n", iso->nVertices);

  /* initialize index pointers */
  for (ii = 0; ii < 3; ii++)
  {
    fPtr[ii] = iso->vertices[ii];
  }

  /* write the vertices */
  for (ii = 0; ii < iso->nVertices; ii++)
  {
    fprintf(f, "    <%g, %g, %g>,\n", *fPtr[0], *fPtr[1], *fPtr[2]);
    fPtr[0]++;
    fPtr[1]++;
    fPtr[2]++;
  }

  fprintf(f, "  }\n");
}

void writeNormals(isosurface *iso, FILE *f)
{

  float *fPtr[3]; /* pointer into the normal list */
  long ii;


  fprintf(f, "  normal_vectors\n  { ");
  fprintf(f, "  %d,\n", iso->nVertices);

  /* initialize index pointers */
  for (ii = 0; ii < 3; ii++)
  {
    fPtr[ii] = iso->normals[ii];
  }

  /* write the normals */
  for (ii = 0; ii < iso->nVertices; ii++)
  {
    fprintf(f, "    <%g, %g, %g>,\n", *fPtr[0], *fPtr[1], *fPtr[2]);
    fPtr[0]++;
    fPtr[1]++;
    fPtr[2]++;
  }

  fprintf(f, "  }\n");
}


void writeTriangles(isosurface *iso, FILE *f)
{


  long *lPtr[3]; /* pointer into the triangle list */
  long ii;


  fprintf(f, "face_indices\n  {");
  fprintf(f, "  %d,\n", iso->nTriangles);

  /* initialize index pointers */
  for (ii = 0; ii < 3; ii++)
  {
    lPtr[ii] = iso->triangles[ii];
  }

  /* write the triangles  */
  for (ii = 0; ii < iso->nTriangles; ii++)
  {
    fprintf(f, "    <%i, %i, %i>,\n", *lPtr[0], *lPtr[1], *lPtr[2]);
    lPtr[0]++;
    lPtr[1]++;
    lPtr[2]++;
  }

  fprintf(f, "  }\n  inside_vector <0, 0, 1>\n}\n");

}

